"""LLM provider configuration utilities."""

from realtimex_toolkit.llm.providers import LLMProviderManager


def configure_provider(provider: str, **kwargs) -> dict[str, str]:
    """Configure LLM provider and set environment variables.

    Convenience function wrapping LLMProviderManager.configure().

    Args:
        provider: Provider name (e.g., 'openai', 'anthropic')
        **kwargs: Additional arguments passed to LLMProviderManager.configure()

    Returns:
        Dictionary of environment variables that were set

    Example:
        >>> configure_provider("openai")
        {'OPENAI_API_KEY': 'sk-...'}
    """
    return LLMProviderManager.configure(provider, **kwargs)


def get_provider_env_vars(provider: str, **kwargs) -> dict[str, str]:
    """Get provider environment variables without setting them.

    Args:
        provider: Provider name (e.g., 'openai', 'anthropic')
        **kwargs: Additional arguments passed to LLMProviderManager constructor

    Returns:
        Dictionary of LiteLLM-compatible environment variables

    Example:
        >>> get_provider_env_vars("anthropic")
        {'ANTHROPIC_API_KEY': 'sk-ant-...'}
    """
    manager = LLMProviderManager(**kwargs)
    return manager.get_provider_env_vars(provider)


__all__ = [
    "LLMProviderManager",
    "configure_provider",
    "get_provider_env_vars",
]



"""Shared pytest fixtures for realtimex_toolkit tests."""

import pytest


@pytest.fixture
def mock_env_file(tmp_path):
    """Create a temporary .env file for testing."""
    env_content = "OPEN_AI_KEY=test-key-from-file\nANTHROPIC_API_KEY=test-anthropic-key\n"
    env_file = tmp_path / ".env.development"
    env_file.write_text(env_content)
    return env_file



"""Credential retrieval and decryption utilities."""

from __future__ import annotations

import asyncio
import base64
import hashlib
import json
import logging
import os
import threading
from typing import Any
from urllib.parse import quote

from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from realtimex_toolkit.api.http_client import ApiClient
from realtimex_toolkit.credentials.models import CredentialBundle, CredentialType
from realtimex_toolkit.exceptions import ApiError, CredentialError
from realtimex_toolkit.utils.path_utils import get_shared_env_path

try:  # pragma: no cover - optional dependency is always available in prod envs
    from dotenv import dotenv_values
except ImportError:  # pragma: no cover - fallback when python-dotenv not installed
    dotenv_values = None  # type: ignore[assignment]

logger = logging.getLogger(__name__)


class _EventLoopThread:
    """Background thread with a long-lived event loop for running async operations."""

    def __init__(self) -> None:
        self._loop: asyncio.AbstractEventLoop | None = None
        self._thread: threading.Thread | None = None
        self._started = threading.Event()

    def start(self) -> None:
        """Start the background thread and event loop."""
        if self._thread is not None:
            return

        self._thread = threading.Thread(target=self._run_event_loop, daemon=True)
        self._thread.start()
        self._started.wait()

    def _run_event_loop(self) -> None:
        """Run the event loop in the background thread."""
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        self._started.set()
        self._loop.run_forever()

    def run_coroutine(self, coro):
        """Schedule a coroutine on the event loop and return the result."""
        if self._loop is None:
            raise RuntimeError("Event loop thread not started")

        future = asyncio.run_coroutine_threadsafe(coro, self._loop)
        return future.result()

    def stop(self) -> None:
        """Stop the event loop and thread."""
        if self._loop is not None and self._thread is not None:
            self._loop.call_soon_threadsafe(self._loop.stop)
            self._thread.join(timeout=5.0)
            self._loop.close()
            self._loop = None
            self._thread = None


class CredentialManager:
    """Loads credential records from the RealTimeX app backend and decrypts them for consumption."""

    DEFAULT_BASE_URL = "http://localhost:3001"

    def __init__(
        self,
        base_url: str = DEFAULT_BASE_URL,
        api_key: str = "",
        *,
        timeout: int = 30,
        max_retries: int = 3,
    ) -> None:
        """Initialize CredentialManager.

        Args:
            base_url: Base URL for the API (default: http://localhost:3001)
            api_key: API authentication key
            timeout: Request timeout in seconds (default: 30)
            max_retries: Maximum number of retry attempts (default: 3)
        """
        self._loop_thread = _EventLoopThread()
        self._loop_thread.start()
        self._closed = False

        self._api_client = ApiClient(
            base_url=base_url,
            api_key=api_key,
            timeout=timeout,
            max_retries=max_retries,
        )
        self._cache: dict[str, CredentialBundle] = {}
        self._cache_lock = threading.Lock()

        key, salt = self._load_key_material()
        self._encryption_key = self._derive_key(key, salt)

    def close(self) -> None:
        """Close the underlying HTTP client session and stop the event loop thread."""
        if self._closed:
            return

        self._closed = True
        try:
            self._loop_thread.run_coroutine(self._api_client.close())
        except RuntimeError:
            # Event loop already torn down - nothing left to close
            pass
        finally:
            with self._cache_lock:
                self._cache.clear()
            self._loop_thread.stop()

    def __enter__(self) -> CredentialManager:
        return self

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        self.close()

    def get(self, credential_id: str, *, force_refresh: bool = False) -> CredentialBundle:
        """Retrieve and decrypt a credential by id.

        Args:
            credential_id: The unique identifier for the credential
            force_refresh: If True, bypass cache and fetch fresh data

        Returns:
            CredentialBundle containing decrypted credential data

        Raises:
            CredentialError: If credential cannot be retrieved or decrypted
        """
        if self._closed:
            raise CredentialError("Credential manager is closed", credential_id=credential_id)

        if not force_refresh:
            cached = self._cache.get(credential_id)
            if cached is not None:
                return cached

        with self._cache_lock:
            if self._closed:
                raise CredentialError("Credential manager is closed", credential_id=credential_id)

            if not force_refresh:
                cached = self._cache.get(credential_id)
                if cached is not None:
                    return cached

            raw_record = self._loop_thread.run_coroutine(self._fetch_credential(credential_id))
            bundle = self._build_bundle(raw_record)
            self._cache[credential_id] = bundle
            return bundle

    def clear_cache(self) -> None:
        """Evict cached credential bundles."""
        with self._cache_lock:
            self._cache.clear()

    async def _fetch_credential(self, credential_id: str) -> dict[str, Any]:
        try:
            encoded_credential_id = quote(str(credential_id), safe="")
            payload = await self._api_client.request(
                "GET", f"/api/v1/credentials/{encoded_credential_id}"
            )
        except ApiError as exc:  # pragma: no cover - network errors exercised in integration tests
            raise CredentialError(
                f"Failed to fetch credential '{credential_id}' from credential service",
                credential_id=credential_id,
            ) from exc

        if not isinstance(payload, dict):
            raise CredentialError(
                "Credential service returned an unexpected response",
                credential_id=credential_id,
            )

        return self._extract_credential(payload, credential_id)

    def _extract_credential(self, payload: dict[str, Any], credential_id: str) -> dict[str, Any]:
        status = payload.get("status")
        if status is False:
            raise CredentialError(
                "Credential service reported failure",
                credential_id=credential_id,
                details={"status": status},
            )

        credential = payload.get("credential")
        if not isinstance(credential, dict):
            raise CredentialError(
                "Credential service returned malformed credential payload",
                credential_id=credential_id,
            )

        return credential

    def _build_bundle(self, record: dict[str, Any]) -> CredentialBundle:
        credential_id = str(record.get("id", ""))
        raw_type = record.get("type")
        if not raw_type:
            raise CredentialError("Credential record missing type", credential_id=credential_id)

        try:
            credential_type = CredentialType(raw_type)
        except ValueError as exc:
            raise CredentialError(
                f"Unsupported credential type '{raw_type}'",
                credential_id=credential_id,
            ) from exc

        encrypted_blob = record.get("data")
        if not isinstance(encrypted_blob, str) or not encrypted_blob:
            raise CredentialError("Credential record missing payload", credential_id=credential_id)

        decrypted_json = self._decrypt_blob(encrypted_blob, credential_id)
        payload = self._parse_payload(decrypted_json, credential_type, credential_id)

        metadata = record.get("metadata") if isinstance(record.get("metadata"), dict) else None
        updated_at = record.get("updated_at") if isinstance(record.get("updated_at"), str) else None
        name = str(record.get("name")) if record.get("name") is not None else credential_id

        return CredentialBundle(
            credential_id=credential_id,
            name=name,
            credential_type=credential_type,
            payload=payload,
            metadata=metadata,
            updated_at=updated_at,
        )

    def _decrypt_blob(self, encoded_blob: str, credential_id: str) -> str:
        try:
            encrypted_bytes = base64.b64decode(encoded_blob)
        except ValueError as exc:
            raise CredentialError(
                "Credential payload is not valid base64",
                credential_id=credential_id,
            ) from exc

        try:
            encrypted_text = encrypted_bytes.decode("utf-8")
        except UnicodeDecodeError as exc:
            raise CredentialError(
                "Credential payload is not UTF-8 encoded",
                credential_id=credential_id,
            ) from exc

        try:
            cipher_hex, iv_hex = encrypted_text.split(":", 1)
        except ValueError as exc:
            raise CredentialError(
                "Credential payload is missing initialization vector",
                credential_id=credential_id,
            ) from exc

        try:
            cipher_bytes = bytes.fromhex(cipher_hex)
            iv_bytes = bytes.fromhex(iv_hex)
        except ValueError as exc:
            raise CredentialError(
                "Credential payload contains invalid hex data",
                credential_id=credential_id,
            ) from exc

        cipher = Cipher(algorithms.AES(self._encryption_key), modes.CBC(iv_bytes))
        decryptor = cipher.decryptor()
        padded_plaintext = decryptor.update(cipher_bytes) + decryptor.finalize()

        unpadder = padding.PKCS7(algorithms.AES.block_size).unpadder()
        try:
            plaintext_bytes = unpadder.update(padded_plaintext) + unpadder.finalize()
        except ValueError as exc:
            raise CredentialError(
                "Credential payload failed padding validation",
                credential_id=credential_id,
            ) from exc

        try:
            return plaintext_bytes.decode("utf-8")
        except UnicodeDecodeError as exc:
            raise CredentialError(
                "Credential payload could not be decoded",
                credential_id=credential_id,
            ) from exc

    def _parse_payload(
        self,
        payload_json: str,
        credential_type: CredentialType,
        credential_id: str,
    ) -> dict[str, str]:
        try:
            payload = json.loads(payload_json)
        except json.JSONDecodeError as exc:
            raise CredentialError(
                "Credential payload is not valid JSON",
                credential_id=credential_id,
            ) from exc

        if not isinstance(payload, dict):
            raise CredentialError(
                "Credential payload must be a JSON object",
                credential_id=credential_id,
            )

        if credential_type is CredentialType.BASIC_AUTH:
            username = payload.get("username")
            password = payload.get("password")
            if not isinstance(username, str) or not isinstance(password, str):
                raise CredentialError(
                    "Basic auth credential payload requires username and password",
                    credential_id=credential_id,
                )
            return {"username": username, "password": password}

        if credential_type in {
            CredentialType.HTTP_HEADER,
            CredentialType.QUERY_AUTH,
            CredentialType.ENV_VAR,
        }:
            name = payload.get("name")
            value = payload.get("value")
            if not isinstance(name, str) or not isinstance(value, str):
                raise CredentialError(
                    "Credential payload requires name and value fields",
                    credential_id=credential_id,
                )
            return {"name": name, "value": value}

        raise CredentialError(
            f"Parsing not implemented for credential type '{credential_type.value}'",
            credential_id=credential_id,
        )

    def _load_key_material(self) -> tuple[str, str]:
        env_path = get_shared_env_path()
        if not os.path.exists(env_path):
            raise CredentialError("Shared credential environment file not found")

        env_values = self._load_env_file(env_path)
        key = env_values.get("SIG_KEY")
        salt = env_values.get("SIG_SALT")

        if not key or not salt:
            raise CredentialError("Missing SIG_KEY or SIG_SALT for credential decryption")

        return key, salt

    def _load_env_file(self, env_path: str) -> dict[str, str]:
        if dotenv_values is None:
            logger.warning(
                "python-dotenv not installed; cannot load shared .env file: %s", env_path
            )
            return {}

        try:
            values = dotenv_values(env_path) or {}
        except Exception as exc:  # pragma: no cover - defensive guard
            logger.warning("Failed to load shared .env file %s: %s", env_path, str(exc))
            return {}

        return {key: value for key, value in values.items() if isinstance(value, str)}

    def _derive_key(self, key: str, salt: str) -> bytes:
        try:
            return hashlib.scrypt(
                key.encode("utf-8"),
                salt=salt.encode("utf-8"),
                n=16384,
                r=8,
                p=1,
                dklen=32,
            )
        except ValueError as exc:
            raise CredentialError("Unable to derive encryption key material") from exc



"""Tests for LLM provider configuration."""

import os
from pathlib import Path

import pytest

from realtimex_toolkit import LLMProviderManager, configure_provider, get_provider_env_vars
from realtimex_toolkit.exceptions import CredentialError


class TestLLMProviderManager:
    """Test LLMProviderManager class."""

    def test_get_credentials_from_config(self):
        """Test that credentials from config are returned as env vars."""
        providers = {"openai": {"OPEN_AI_KEY": "test-config-key"}}
        manager = LLMProviderManager(providers=providers, shared_env_path="")

        env_vars = manager.get_provider_env_vars("openai")

        assert env_vars == {"OPENAI_API_KEY": "test-config-key"}

    def test_fallback_to_env_file(self, mock_env_file):
        """Provider not in config should fall back to shared .env file."""
        manager = LLMProviderManager(providers={}, shared_env_path=mock_env_file)
        env_vars = manager.get_provider_env_vars("openai")

        assert env_vars == {"OPENAI_API_KEY": "test-key-from-file"}

    def test_config_overrides_env_file(self, mock_env_file):
        """Configured credentials take precedence over shared .env file."""
        providers = {"openai": {"OPEN_AI_KEY": "test-config-key"}}
        manager = LLMProviderManager(providers=providers, shared_env_path=mock_env_file)

        env_vars = manager.get_provider_env_vars("openai")

        assert env_vars == {"OPENAI_API_KEY": "test-config-key"}

    def test_missing_credentials_raises_error(self):
        """Missing config and shared file should raise CredentialError."""
        manager = LLMProviderManager(providers={}, shared_env_path="")

        with pytest.raises(
            CredentialError,
            match="Missing required credential 'OPEN_AI_KEY' for provider 'openai'",
        ):
            manager.get_provider_env_vars("openai")

    def test_unsupported_provider_raises_error(self):
        """Unsupported provider should raise ValueError."""
        manager = LLMProviderManager(providers={}, shared_env_path="")

        with pytest.raises(ValueError, match="Unsupported provider: 'unknown-provider'"):
            manager.get_provider_env_vars("unknown-provider")

    def test_configure_provider_sets_env_vars(self):
        """Test that configure_provider correctly sets environment variables."""
        providers = {"openai": {"OPEN_AI_KEY": "test-env-key"}}
        manager = LLMProviderManager(providers=providers, shared_env_path="")

        result = manager.configure_provider("openai")

        assert os.environ.get("OPENAI_API_KEY") == "test-env-key"
        assert result == {"OPENAI_API_KEY": "test-env-key"}

        del os.environ["OPENAI_API_KEY"]

    def test_multi_credential_provider(self):
        """Test provider requiring multiple credentials (Azure)."""
        providers = {
            "azure": {
                "AZURE_OPENAI_KEY": "test-azure-key",
                "AZURE_OPENAI_ENDPOINT": "https://test.openai.azure.com",
            }
        }
        manager = LLMProviderManager(providers=providers, shared_env_path="")

        env_vars = manager.get_provider_env_vars("azure")

        assert env_vars == {
            "AZURE_OPENAI_API_KEY": "test-azure-key",
            "AZURE_OPENAI_ENDPOINT": "https://test.openai.azure.com",
        }

    def test_optional_credentials(self):
        """Test providers with optional credentials (localai)."""
        providers = {"localai": {"LOCAL_AI_BASE_PATH": "http://localhost:8080"}}
        manager = LLMProviderManager(providers=providers, shared_env_path="")

        env_vars = manager.get_provider_env_vars("localai")

        assert env_vars == {"OPENAI_API_BASE": "http://localhost:8080"}


class TestClassmethodConfigure:
    """Test the classmethod configure() for utility-style usage."""

    def test_configure_with_explicit_providers(self):
        """Test configure() classmethod with explicit providers."""
        providers = {"openai": {"OPEN_AI_KEY": "test-classmethod-key"}}
        mock_setter = {}

        def setter(key, value):
            mock_setter[key] = value

        result = LLMProviderManager.configure(
            "openai",
            providers=providers,
            shared_env_path="",
            env_setter=setter,
        )

        assert result == {"OPENAI_API_KEY": "test-classmethod-key"}
        assert mock_setter == {"OPENAI_API_KEY": "test-classmethod-key"}

    def test_configure_with_shared_env_file(self, mock_env_file):
        """Test configure() classmethod falls back to shared env file."""
        mock_setter = {}

        def setter(key, value):
            mock_setter[key] = value

        result = LLMProviderManager.configure(
            "anthropic",
            providers={},
            shared_env_path=mock_env_file,
            env_setter=setter,
        )

        assert result == {"ANTHROPIC_API_KEY": "test-anthropic-key"}
        assert mock_setter == {"ANTHROPIC_API_KEY": "test-anthropic-key"}


class TestConvenienceFunctions:
    """Test top-level convenience functions."""

    def test_configure_provider_function(self):
        """Test configure_provider() convenience function."""
        providers = {"openai": {"OPEN_AI_KEY": "test-convenience-key"}}
        mock_setter = {}

        def setter(key, value):
            mock_setter[key] = value

        result = configure_provider(
            "openai",
            providers=providers,
            shared_env_path="",
            env_setter=setter,
        )

        assert result == {"OPENAI_API_KEY": "test-convenience-key"}
        assert mock_setter == {"OPENAI_API_KEY": "test-convenience-key"}

    def test_get_provider_env_vars_function(self):
        """Test get_provider_env_vars() convenience function."""
        providers = {"openai": {"OPEN_AI_KEY": "test-get-key"}}

        env_vars = get_provider_env_vars("openai", providers=providers, shared_env_path="")

        assert env_vars == {"OPENAI_API_KEY": "test-get-key"}


class TestPathHandling:
    """Test path handling for shared env file."""

    def test_path_object_as_shared_env_path(self, mock_env_file):
        """Test that Path objects are accepted for shared_env_path."""
        manager = LLMProviderManager(providers={}, shared_env_path=Path(mock_env_file))
        env_vars = manager.get_provider_env_vars("openai")

        assert env_vars == {"OPENAI_API_KEY": "test-key-from-file"}

    def test_string_path_as_shared_env_path(self, mock_env_file):
        """Test that string paths are accepted for shared_env_path."""
        manager = LLMProviderManager(providers={}, shared_env_path=str(mock_env_file))
        env_vars = manager.get_provider_env_vars("openai")

        assert env_vars == {"OPENAI_API_KEY": "test-key-from-file"}



"""HTTP API client utilities."""

from realtimex_toolkit.api.error_mapping import DefaultErrorMapper, ErrorMapper
from realtimex_toolkit.api.http_client import ApiClient

__all__ = [
    "ApiClient",
    "DefaultErrorMapper",
    "ErrorMapper",
]



"""Internal utilities for RealtimeX."""

from realtimex_toolkit.utils.path_utils import get_realtimex_user_dir, get_shared_env_path

__all__ = [
    "get_realtimex_user_dir",
    "get_shared_env_path",
]



"""RealtimeX Internal Utilities - Lightweight library for LLM and credential management."""

__version__ = "1.1.0"

from realtimex_toolkit.credentials import (
    CredentialBundle,
    CredentialManager,
    CredentialType,
    get_credential,
)

from realtimex_toolkit.agent_flow import (
    get_flow_variable,
    get_thread_id,
    get_workspace_slug,
    get_workspace_data_dir
)

from realtimex_toolkit.exceptions import (
    ApiError,
    AuthenticationError,
    ConnectionError,
    CredentialError,
    ProviderError,
    RateLimitError,
    RealtimeXError,
    ResourceNotFoundError,
    ServerError,
)
from realtimex_toolkit.llm import LLMProviderManager, configure_provider, get_provider_env_vars

__all__ = [
    # Version
    "__version__",
    # Exceptions
    "ApiError",
    "AuthenticationError",
    "ConnectionError",
    "CredentialError",
    "ProviderError",
    "RateLimitError",
    "RealtimeXError",
    "ResourceNotFoundError",
    "ServerError",
    # LLM
    "LLMProviderManager",
    "configure_provider",
    "get_provider_env_vars",
    # Credentials
    "CredentialBundle",
    "CredentialManager",
    "CredentialType",
    "get_credential",
    "get_flow_variable",
    "get_thread_id",
    "get_workspace_slug",
    "get_workspace_data_dir"
]



"""Custom exception classes for RealtimeX utilities."""

from typing import Any


class RealtimeXError(Exception):
    """Base exception for all RealtimeX errors."""

    def __init__(self, message: str = "") -> None:
        super().__init__(message)

    @property
    def message(self) -> str:
        return self.args[0] if self.args else ""

    def __str__(self) -> str:  # pragma: no cover - default formatting
        return self.message or super().__str__()


class ApiError(RealtimeXError):
    """Raised when API communication fails."""

    def __init__(
        self,
        message: str,
        status_code: int | None = None,
        response_body: str | None = None,
    ) -> None:
        self.status_code = status_code
        self.response_body = response_body
        super().__init__(message)

    def __str__(self) -> str:
        message = self.message or "API error"
        if self.status_code is not None:
            return f"{message} | Status: {self.status_code}"
        return message


class ConnectionError(ApiError):
    """Raised when network connection fails."""

    pass


class ResourceNotFoundError(ApiError):
    """Raised when a requested resource cannot be located (HTTP 404)."""

    def __init__(
        self,
        message: str = "Resource not found",
        status_code: int | None = 404,
        response_body: str | None = None,
    ) -> None:
        super().__init__(message, status_code=status_code, response_body=response_body)


class AuthenticationError(ApiError):
    """Raised when API authentication fails."""

    def __init__(
        self,
        message: str = "Authentication failed",
        status_code: int | None = None,
        response_body: str | None = None,
    ) -> None:
        super().__init__(message, status_code=status_code, response_body=response_body)


class RateLimitError(ApiError):
    """Raised when API rate limit is exceeded."""

    def __init__(
        self,
        message: str = "Rate limit exceeded",
        status_code: int | None = None,
        response_body: str | None = None,
        retry_after: int | None = None,
    ) -> None:
        super().__init__(message, status_code=status_code, response_body=response_body)
        self.retry_after = retry_after


class ServerError(ApiError):
    """Raised when server returns 5xx error."""

    def __init__(
        self,
        message: str = "Server error",
        status_code: int | None = None,
        response_body: str | None = None,
    ) -> None:
        super().__init__(message, status_code=status_code, response_body=response_body)


class CredentialError(RealtimeXError):
    """Raised when credential retrieval or decryption fails."""

    def __init__(
        self,
        message: str,
        *,
        credential_id: str | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        self.credential_id = credential_id
        self.details = details or {}
        super().__init__(message)

    def __str__(self) -> str:
        base = self.message or "Credential error"
        extras: list[str] = []
        if self.credential_id:
            extras.append(f"id={self.credential_id}")
        if self.details:
            extras.append(f"details={self.details}")
        if extras:
            return f"{base} ({', '.join(extras)})"
        return base


class ProviderError(RealtimeXError):
    """Raised when LLM provider configuration fails."""

    pass



"""Credential management utilities."""

from typing import Any

from realtimex_toolkit.credentials.manager import CredentialManager
from realtimex_toolkit.credentials.models import CredentialBundle, CredentialType


def get_credential(
    credential_id: str,
    *,
    api_key: str = "",
    base_url: str | None = None,
    force_refresh: bool = False,
) -> dict[str, Any]:
    """Retrieve and decrypt a credential by ID - convenience wrapper.

    This is a convenience function that creates a CredentialManager instance,
    fetches the credential, and properly closes the connection.

    Args:
        credential_id: The unique identifier for the credential
        api_key: API authentication key (default: empty string)
        base_url: Base URL for the API (default: http://localhost:3001)
        force_refresh: If True, bypass cache and fetch fresh data (default: False)

    Returns:
        Dictionary containing decrypted credential data in a JSON-friendly shape:
        {
            "credential_id": str,
            "name": str,
            "credential_type": str,
            "payload": dict[str, str],
            "metadata": dict | None,
            "updated_at": str | None,
        }

    Raises:
        CredentialError: If credential cannot be retrieved or decrypted

    Example:
        >>> bundle = get_credential("cred-123", api_key="your-api-key")
        >>> print(bundle["payload"])
        {'name': 'API_KEY', 'value': 'secret-value'}
    """
    kwargs = {"api_key": api_key}
    if base_url is not None:
        kwargs["base_url"] = base_url

    manager = CredentialManager(**kwargs)
    try:
        bundle = manager.get(credential_id, force_refresh=force_refresh)
        return bundle.to_dict()
    finally:
        manager.close()


__all__ = [
    "CredentialBundle",
    "CredentialManager",
    "CredentialType",
    "get_credential",
]



"""LLM provider credential mappings for LiteLLM compatibility."""

PROVIDER_CREDENTIAL_MAPPINGS: dict[str, list[tuple[str, str, bool]]] = {
    "openai": [
        ("OPEN_AI_KEY", "OPENAI_API_KEY", True),
    ],
    "anthropic": [
        ("ANTHROPIC_API_KEY", "ANTHROPIC_API_KEY", True),
    ],
    "azure": [
        ("AZURE_OPENAI_KEY", "AZURE_OPENAI_API_KEY", True),
        ("AZURE_OPENAI_ENDPOINT", "AZURE_OPENAI_ENDPOINT", True),
    ],
    "gemini": [
        ("GEMINI_API_KEY", "GEMINI_API_KEY", True),
    ],
    "realtimexai": [
        ("REALTIMEX_AI_API_KEY", "OPENAI_API_KEY", True),
        ("REALTIMEX_AI_BASE_PATH", "OPENAI_API_BASE", True),
    ],
    "ollama": [
        ("OLLAMA_BASE_PATH", "OPENAI_API_BASE", True),
    ],
    "groq": [
        ("GROQ_API_KEY", "GROQ_API_KEY", True),
    ],
    "cohere": [
        ("COHERE_API_KEY", "COHERE_API_KEY", True),
    ],
    "mistral": [
        ("MISTRAL_API_KEY", "MISTRAL_API_KEY", True),
    ],
    "perplexity": [
        ("PERPLEXITY_API_KEY", "PERPLEXITYAI_API_KEY", True),
    ],
    "openrouter": [
        ("OPENROUTER_API_KEY", "OPENROUTER_API_KEY", True),
    ],
    "togetherai": [
        ("TOGETHER_AI_API_KEY", "TOGETHERAI_API_KEY", True),
    ],
    "fireworksai": [
        ("FIREWORKS_AI_LLM_API_KEY", "FIREWORKS_API_KEY", True),
    ],
    "deepseek": [
        ("DEEPSEEK_API_KEY", "DEEPSEEK_API_KEY", True),
    ],
    "xai": [
        ("XAI_LLM_API_KEY", "XAI_API_KEY", True),
    ],
    "novita": [
        ("NOVITA_LLM_API_KEY", "NOVITA_API_KEY", True),
    ],
    "bedrock": [
        ("AWS_BEDROCK_LLM_ACCESS_KEY_ID", "AWS_ACCESS_KEY_ID", True),
        ("AWS_BEDROCK_LLM_ACCESS_KEY", "AWS_SECRET_ACCESS_KEY", True),
        ("AWS_BEDROCK_LLM_REGION", "AWS_REGION_NAME", True),
    ],
    "localai": [
        ("LOCAL_AI_BASE_PATH", "OPENAI_API_BASE", True),
        ("LOCAL_AI_API_KEY", "OPENAI_API_KEY", False),
    ],
    "lmstudio": [
        ("LMSTUDIO_BASE_PATH", "OPENAI_API_BASE", True),
    ],
    "textgenwebui": [
        ("TEXT_GEN_WEB_UI_BASE_PATH", "OPENAI_API_BASE", True),
        ("TEXT_GEN_WEB_UI_API_KEY", "OPENAI_API_KEY", False),
    ],
    "koboldcpp": [
        ("KOBOLD_CPP_BASE_PATH", "OPENAI_API_BASE", True),
    ],
    "litellm": [
        ("LITE_LLM_BASE_PATH", "OPENAI_API_BASE", True),
        ("LITE_LLM_API_KEY", "OPENAI_API_KEY", False),
    ],
    "generic-openai": [
        ("GENERIC_OPEN_AI_BASE_PATH", "OPENAI_API_BASE", True),
        ("GENERIC_OPEN_AI_API_KEY", "OPENAI_API_KEY", True),
    ],
    "nvidia-nim": [
        ("NVIDIA_NIM_LLM_BASE_PATH", "OPENAI_API_BASE", True),
    ],
    "huggingface": [
        ("HUGGING_FACE_LLM_API_KEY", "HUGGINGFACE_API_KEY", True),
        ("HUGGING_FACE_LLM_ENDPOINT", "OPENAI_API_BASE", True),
    ],
    "dpais": [
        ("DPAIS_LLM_BASE_PATH", "OPENAI_API_BASE", True),
    ],
    "apipie": [
        ("APIPIE_LLM_API_KEY", "APIPIE_API_KEY", True),
    ],
    "ppio": [
        ("PPIO_API_KEY", "PPIO_API_KEY", True),
    ],
}



"""Tests for Credential Manager."""

import base64
import hashlib
import json
from unittest.mock import AsyncMock, patch
from urllib.parse import quote

import pytest
import pytest_asyncio

from realtimex_toolkit import CredentialBundle, CredentialManager, CredentialType
from realtimex_toolkit.exceptions import ApiError, CredentialError


class TestCredentialManager:
    """Test CredentialManager class."""

    @pytest.fixture
    def mock_env_file_with_sig(self, tmp_path):
        """Create a temporary .env file with signature keys for testing."""
        env_content = (
            "SIG_KEY=test-signature-key\nSIG_SALT=test-signature-salt\nOPEN_AI_KEY=test-key\n"
        )
        env_file = tmp_path / ".env.development"
        env_file.write_text(env_content)
        return env_file

    @pytest.fixture
    def encrypted_credential(self):
        """Create an encrypted credential for testing."""
        # Use the same keys as in mock_env_file_with_sig
        key = "test-signature-key"
        salt = "test-signature-salt"

        # Derive encryption key
        encryption_key = hashlib.scrypt(
            key.encode("utf-8"),
            salt=salt.encode("utf-8"),
            n=16384,
            r=8,
            p=1,
            dklen=32,
        )

        # Create credential payload
        payload = {"name": "API_KEY", "value": "secret-value-123"}
        payload_json = json.dumps(payload)

        # Encrypt using AES-CBC (simplified for test)
        from cryptography.hazmat.primitives import padding
        from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

        iv = b"0123456789abcdef"  # 16 bytes IV
        cipher = Cipher(algorithms.AES(encryption_key), modes.CBC(iv))
        encryptor = cipher.encryptor()

        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(payload_json.encode("utf-8")) + padder.finalize()
        ciphertext = encryptor.update(padded_data) + encryptor.finalize()

        # Format as expected by _decrypt_blob
        cipher_hex = ciphertext.hex()
        iv_hex = iv.hex()
        encrypted_text = f"{cipher_hex}:{iv_hex}"
        encoded_blob = base64.b64encode(encrypted_text.encode("utf-8")).decode("utf-8")

        return encoded_blob, payload

    @pytest_asyncio.fixture
    async def manager(self, mock_env_file_with_sig):
        """Create a CredentialManager instance for testing."""
        with patch("realtimex_toolkit.credentials.manager.get_shared_env_path") as mock_path:
            mock_path.return_value = str(mock_env_file_with_sig)
            manager = CredentialManager(
                base_url="https://api.example.com",
                api_key="test-key",
            )
            yield manager
            manager.close()

    @pytest.mark.asyncio
    async def test_get_credential_success(self, manager, encrypted_credential):
        """Test successful credential retrieval and decryption."""
        encoded_blob, expected_payload = encrypted_credential

        api_response = {
            "status": True,
            "credential": {
                "id": "cred-123",
                "name": "Test Credential",
                "type": "env_var",
                "data": encoded_blob,
                "metadata": {"tags": ["production"]},
                "updated_at": "2025-01-01T00:00:00Z",
            },
        }

        manager._api_client.request = AsyncMock(return_value=api_response)

        result = manager.get("cred-123")

        assert isinstance(result, CredentialBundle)
        assert result.credential_id == "cred-123"
        assert result.name == "Test Credential"
        assert result.credential_type == CredentialType.ENV_VAR
        assert result.payload == expected_payload
        assert result.metadata == {"tags": ["production"]}
        assert result.updated_at == "2025-01-01T00:00:00Z"

    @pytest.mark.asyncio
    async def test_get_credential_caching(self, manager, encrypted_credential):
        """Test that credentials are cached after first retrieval."""
        encoded_blob, _ = encrypted_credential

        api_response = {
            "status": True,
            "credential": {
                "id": "cred-cached",
                "name": "Cached Credential",
                "type": "http_header",
                "data": encoded_blob,
            },
        }

        manager._api_client.request = AsyncMock(return_value=api_response)

        # First call - should fetch from API
        result1 = manager.get("cred-cached")
        assert manager._api_client.request.call_count == 1

        # Second call - should use cache
        result2 = manager.get("cred-cached")
        assert manager._api_client.request.call_count == 1  # No additional call
        assert result1 is result2

    @pytest.mark.asyncio
    async def test_get_credential_encodes_identifier(self, manager, encrypted_credential):
        """Ensure credential identifiers are safely encoded before request."""
        encoded_blob, _ = encrypted_credential
        special_id = "name with spaces/#hash"

        api_response = {
            "status": True,
            "credential": {
                "id": "cred-special",
                "name": "Special Credential",
                "type": "env_var",
                "data": encoded_blob,
            },
        }

        manager._api_client.request = AsyncMock(return_value=api_response)

        manager.get(special_id)

        method, path = manager._api_client.request.await_args.args
        assert method == "GET"
        assert path == f"/api/v1/credentials/{quote(special_id, safe='')}"

    @pytest.mark.asyncio
    async def test_get_credential_force_refresh(self, manager, encrypted_credential):
        """Test force_refresh bypasses cache."""
        encoded_blob, _ = encrypted_credential

        api_response = {
            "status": True,
            "credential": {
                "id": "cred-refresh",
                "name": "Refresh Credential",
                "type": "basic_auth",
                "data": encoded_blob,
            },
        }

        # Create different encrypted payloads for username/password
        payload = {"username": "testuser", "password": "testpass"}
        payload_json = json.dumps(payload)

        from cryptography.hazmat.primitives import padding
        from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

        encryption_key = manager._encryption_key
        iv = b"0123456789abcdef"
        cipher = Cipher(algorithms.AES(encryption_key), modes.CBC(iv))
        encryptor = cipher.encryptor()

        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(payload_json.encode("utf-8")) + padder.finalize()
        ciphertext = encryptor.update(padded_data) + encryptor.finalize()

        encrypted_text = f"{ciphertext.hex()}:{iv.hex()}"
        api_response["credential"]["data"] = base64.b64encode(
            encrypted_text.encode("utf-8")
        ).decode("utf-8")

        manager._api_client.request = AsyncMock(return_value=api_response)

        # First call
        manager.get("cred-refresh")
        assert manager._api_client.request.call_count == 1

        # Force refresh - should fetch again
        manager.get("cred-refresh", force_refresh=True)
        assert manager._api_client.request.call_count == 2

    @pytest.mark.asyncio
    async def test_get_credential_api_error(self, manager):
        """Test handling of API errors during credential fetch."""
        manager._api_client.request = AsyncMock(
            side_effect=ApiError("Network error", status_code=500)
        )

        with pytest.raises(CredentialError, match="Failed to fetch credential"):
            manager.get("cred-error")

    @pytest.mark.asyncio
    async def test_get_credential_status_false(self, manager):
        """Test handling of failure status in API response."""
        api_response = {"status": False}

        manager._api_client.request = AsyncMock(return_value=api_response)

        with pytest.raises(CredentialError, match="Credential service reported failure"):
            manager.get("cred-fail")

    @pytest.mark.asyncio
    async def test_get_after_close_raises(self, manager):
        """Ensure manager refuses operations once closed."""
        manager.close()
        manager.close()  # verify idempotent

        with pytest.raises(CredentialError, match="manager is closed"):
            manager.get("cred-closed")

    @pytest.mark.asyncio
    async def test_get_credential_malformed_response(self, manager):
        """Test handling of malformed API response."""
        api_response = {"status": True, "credential": "not-a-dict"}

        manager._api_client.request = AsyncMock(return_value=api_response)

        with pytest.raises(CredentialError, match="malformed credential payload"):
            manager.get("cred-malformed")

    @pytest.mark.asyncio
    async def test_unsupported_credential_type(self, manager, encrypted_credential):
        """Test handling of unsupported credential types."""
        encoded_blob, _ = encrypted_credential

        api_response = {
            "status": True,
            "credential": {
                "id": "cred-unsupported",
                "type": "unsupported_type",
                "data": encoded_blob,
            },
        }

        manager._api_client.request = AsyncMock(return_value=api_response)

        with pytest.raises(CredentialError, match="Unsupported credential type"):
            manager.get("cred-unsupported")

    @pytest.mark.asyncio
    async def test_invalid_encrypted_data(self, manager):
        """Test handling of invalid base64 encrypted data."""
        api_response = {
            "status": True,
            "credential": {
                "id": "cred-invalid",
                "type": "env_var",
                "data": "not-valid-base64!!!",
            },
        }

        manager._api_client.request = AsyncMock(return_value=api_response)

        with pytest.raises(CredentialError, match="not valid base64"):
            manager.get("cred-invalid")

    @pytest.mark.asyncio
    async def test_clear_cache(self, manager):
        """Test cache clearing."""
        # Manually populate cache
        bundle = CredentialBundle(
            credential_id="test",
            name="Test",
            credential_type=CredentialType.ENV_VAR,
            payload={"key": "value"},
        )
        manager._cache["test"] = bundle

        assert "test" in manager._cache

        manager.clear_cache()

        assert len(manager._cache) == 0

    def test_credential_manager_init_missing_sig_keys(self, tmp_path):
        """Test CredentialManager initialization fails without signature keys."""
        env_file = tmp_path / ".env.development"
        env_file.write_text("SOME_KEY=value\n")

        with patch("realtimex_toolkit.credentials.manager.get_shared_env_path") as mock_path:
            mock_path.return_value = str(env_file)

            with pytest.raises(CredentialError, match="Missing SIG_KEY or SIG_SALT"):
                CredentialManager(base_url="https://api.example.com", api_key="test-key")

    def test_credential_manager_init_no_env_file(self):
        """Test CredentialManager initialization fails when env file doesn't exist."""
        with patch("realtimex_toolkit.credentials.manager.get_shared_env_path") as mock_path:
            mock_path.return_value = "/nonexistent/path/.env.development"

            with pytest.raises(CredentialError, match="environment file not found"):
                CredentialManager(base_url="https://api.example.com", api_key="test-key")


class TestGetCredentialConvenience:
    """Test get_credential convenience function."""

    @pytest.fixture
    def mock_env_file_with_sig(self, tmp_path):
        """Create a temporary .env file with signature keys for testing."""
        env_content = (
            "SIG_KEY=test-signature-key\nSIG_SALT=test-signature-salt\nOPEN_AI_KEY=test-key\n"
        )
        env_file = tmp_path / ".env.development"
        env_file.write_text(env_content)
        return env_file

    @pytest.fixture
    def encrypted_credential(self):
        """Create an encrypted credential for testing."""
        key = "test-signature-key"
        salt = "test-signature-salt"

        encryption_key = hashlib.scrypt(
            key.encode("utf-8"),
            salt=salt.encode("utf-8"),
            n=16384,
            r=8,
            p=1,
            dklen=32,
        )

        payload = {"name": "API_KEY", "value": "secret-value-123"}
        payload_json = json.dumps(payload)

        from cryptography.hazmat.primitives import padding
        from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

        iv = b"0123456789abcdef"
        cipher = Cipher(algorithms.AES(encryption_key), modes.CBC(iv))
        encryptor = cipher.encryptor()

        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(payload_json.encode("utf-8")) + padder.finalize()
        ciphertext = encryptor.update(padded_data) + encryptor.finalize()

        cipher_hex = ciphertext.hex()
        iv_hex = iv.hex()
        encrypted_text = f"{cipher_hex}:{iv_hex}"
        encoded_blob = base64.b64encode(encrypted_text.encode("utf-8")).decode("utf-8")

        return encoded_blob, payload

    @pytest.mark.asyncio
    async def test_get_credential_success(self, mock_env_file_with_sig, encrypted_credential):
        """Test successful credential retrieval using convenience function."""
        from realtimex_toolkit.credentials import get_credential

        encoded_blob, expected_payload = encrypted_credential

        api_response = {
            "status": True,
            "credential": {
                "id": "cred-456",
                "name": "Test Credential",
                "type": "env_var",
                "data": encoded_blob,
            },
        }

        with patch("realtimex_toolkit.credentials.manager.get_shared_env_path") as mock_path:
            mock_path.return_value = str(mock_env_file_with_sig)

            with patch(
                "realtimex_toolkit.api.http_client.ApiClient.request", new_callable=AsyncMock
            ) as mock_request:
                mock_request.return_value = api_response

                result = get_credential("cred-456", api_key="test-key")

                assert isinstance(result, dict)
                assert result["credential_id"] == "cred-456"
                assert result["name"] == "Test Credential"
                assert result["payload"] == expected_payload

    @pytest.mark.asyncio
    async def test_get_credential_with_custom_base_url(
        self, mock_env_file_with_sig, encrypted_credential
    ):
        """Test convenience function with custom base URL."""
        from realtimex_toolkit.credentials import get_credential

        encoded_blob, _ = encrypted_credential

        api_response = {
            "status": True,
            "credential": {
                "id": "cred-789",
                "type": "http_header",
                "data": encoded_blob,
            },
        }

        with patch("realtimex_toolkit.credentials.manager.get_shared_env_path") as mock_path:
            mock_path.return_value = str(mock_env_file_with_sig)

            with patch(
                "realtimex_toolkit.api.http_client.ApiClient.request", new_callable=AsyncMock
            ) as mock_request:
                mock_request.return_value = api_response

                result = get_credential(
                    "cred-456", api_key="test-key", base_url="https://custom-api.example.com"
                )

                assert isinstance(result, dict)
                assert result["credential_id"] == "cred-789"


class TestCredentialBundle:
    """Test CredentialBundle model."""

    def test_as_dict(self):
        """Test as_dict returns a copy of the payload."""
        bundle = CredentialBundle(
            credential_id="test-id",
            name="Test",
            credential_type=CredentialType.HTTP_HEADER,
            payload={"header": "Bearer token"},
        )

        result = bundle.as_dict()

        assert result == {"header": "Bearer token"}
        assert result is not bundle.payload  # Ensure it's a copy

    def test_masked_payload(self):
        """Test masked_payload masks all values."""
        bundle = CredentialBundle(
            credential_id="test-id",
            name="Test",
            credential_type=CredentialType.BASIC_AUTH,
            payload={"username": "user", "password": "secret"},
        )

        result = bundle.masked_payload()

        assert result == {"username": "[MASKED]", "password": "[MASKED]"}

    def test_credential_type_enum(self):
        """Test CredentialType enum values."""
        assert CredentialType.HTTP_HEADER == "http_header"
        assert CredentialType.QUERY_AUTH == "query_auth"
        assert CredentialType.BASIC_AUTH == "basic_auth"
        assert CredentialType.ENV_VAR == "env_var"



"""Path-related utility functions for RealtimeX ecosystem."""

import os


def get_realtimex_user_dir() -> str:
    """Returns the path to the .realtimex.ai user directory.

    Returns:
        Path to the user directory (e.g., ~/.realtimex.ai)
    """
    return os.path.join(os.path.expanduser("~"), ".realtimex.ai")


def get_shared_env_path() -> str:
    """Returns the fixed path to the shared environment file.

    Returns:
        Path to the shared .env.development file
    """
    user_dir = get_realtimex_user_dir()
    return os.path.join(user_dir, "Resources", "server", ".env.development")



"""Agent flow management utilities."""
import os
import sys
import json
from typing import Any

def get_flow_variable(
    variable_name:str = None,
    default_value:Any = None
) -> Any:
    """Retrieve flow variable"""
    try:
        run_id = sys.argv[1]
        payload_file_path = sys.argv[2]
        payload = None

        if not os.path.exists(payload_file_path):
            return None

        with open(payload_file_path, 'r') as f:
            payload = json.load(f)

        if not payload:
            return None

        if variable_name:
            if variable_name in payload:
                return payload[variable_name]
            return default_value
            
        return payload
        
    except Exception:
        return None


def get_workspace_slug(
    default_value:Any = None
) -> Any:
    """Retrieve current workspace slug"""
    try:
        return get_flow_variable("workspace_slug",default_value=default_value)
    except Exception:
        return default_value

def get_thread_id(
    default_value:Any = None
) -> Any:
    """Retrieve current thread id"""
    try:
        return get_flow_variable("thread_id",default_value=default_value)
    except Exception:
        return default_value

def get_workspace_data_dir(
    variable_name:str = None,
    default_workspace_slug:Any = None
) -> Any:
    """Retrieve flow variable"""
    try:
        import os

        workspace_slug = get_workspace_slug(default_value=default_workspace_slug)
        realtimex_dir = os.path.join(os.path.expanduser("~"),".realtimex.ai")
        realtimex_storage_dir = os.path.realpath(os.path.join(realtimex_dir,"Resources","server","storage"))

        if not os.path.exists(realtimex_storage_dir):
            return None

        workspace_data_dir = os.path.join(realtimex_storage_dir,"working-data",workspace_slug)

        if not os.path.exists(workspace_data_dir):
            os.makedirs(workspace_data_dir,exist_ok=True)
        
        return workspace_data_dir

        
    except Exception:
        return None


__all__ = [
    "get_flow_variable",
    "get_workspace_data_dir",
    "get_thread_id",
    "get_workspace_slug"
]
